// test4voids.cxx -- check for voids in an array file, or folder
//
// fillvoids.cxx - Written by Curtis Olson, started November 2005.
//
// Copyright (C) 2005  Curtis L. Olson  - http://www.flightgear.org/~curt
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: test4voids.cxx,v 1.0 2009-02-20 16:30:10 geoff Exp $

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <simgear/compiler.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <string>
#include <iostream>

#include <simgear/bucket/newbucket.hxx>
#include <simgear/debug/logstream.hxx>
#include <simgear/misc/sg_path.hxx>

#include <Array/array.hxx>
#ifdef _MSC_VER
#include <io.h>
#endif
#include <map>

using std::cout;
using std::endl;
using std::string;
using std::map;

static int verbose = 0;

//typedef std::map<string, int> StrMap;
typedef map<string, int> StrMap;
typedef vector<string> string_list;

static StrMap files_with_voids;

static void add_to_missing( string file, int cnt )
{
   files_with_voids[file] = cnt;
}
static void show_missing( void )
{
   cout << "Summary of files with voids..." << endl;
   for( StrMap::iterator it = files_with_voids.begin();
      it != files_with_voids.end(); it++ )
   {
      string file = it->first;
      int cnt = it->second;
      cout << file << " voids " << cnt << endl;
   }
}

static int is_file_or_directory( string cur_item )
{
   struct stat buf;
   if( stat( cur_item.c_str(), &buf ) == 0 ) {
      if ( buf.st_mode & _S_IFDIR )
         return 1; // is directory
      else
         return 2; // is file
   }
   return 0;
}

void give_help( char * name )
{
   cout << "Usage: " << name << " <src_array_dir> | <src_array_file>"
             << endl;
   cout << "Given a single arr.gz file, check it for voids." << endl;
   cout << "Given a directory, will walk the tree seeking *.arr, or *.arr.gz" << endl;
   cout << "files, and list those that contain voids. A void is defined" << endl;
   cout << "as slots with less than -9000 meters." << endl;
}

void check_args( int argc, char **argv )
{
   for(int i = 1; i < argc; i++)
   {
      char * parg = argv[i];
      if(( strcmp(parg,"-h") == 0 )||
         ( strcmp(parg,"--help") == 0 ))
      {
         give_help( argv[0] );
         exit(0);
      }
   }
}

static bool process_array_file( string src_array_path )
{
   cout << "Processing [" << src_array_path << "]" << endl;
   // compute the fill array path
   SGPath tmp1( src_array_path );
   string tmp2 = tmp1.base(); // strip .gz
   //cout << "tmp2 = " << tmp2 << endl;
   string tmp3 = SGPath(tmp2).base(); // strip .arr
   //cout << "tmp3 = " << tmp3 << endl;
   string file = SGPath(tmp3).file();
   //cout << "file = " << file << endl;
   long int index = atoi(file.c_str());
   SGBucket bucket( index );
   if( verbose )
      cout << "Index = " << index << endl;

   // open the source array
   TGArray src_array;
   src_array.open( tmp3 );
   src_array.parse( bucket );
   if ( !src_array.is_open() ) {
     cout << "Unable to open source array " << tmp3 << endl;
     cout << "tmp2 = " << tmp2 ;
     cout << " Bucket Index = " << index << endl;
     cout << "input_file = " << src_array_path << endl;
     return false;
   }
   // traverse the source array and lookup replacement values for any voids
   bool has_void = false;
   long void_count = 0;
   for ( int i = 0; i < src_array.get_cols(); ++i ) {
      for ( int j = 0; j < src_array.get_rows(); ++j ) {
         int src_elev = src_array.get_array_elev(i, j);
         if ( src_elev < -9000 ) {
            has_void = true;
            if( verbose )
               cout << "We have a void = " << src_elev << endl;
            // break;
            void_count++;
	      }
      }
   }
   if(has_void) {
      cout << "File [" << src_array_path << "] has " << void_count << " voids." << endl;
      add_to_missing( src_array_path, void_count );
   }
   src_array.close();

   return has_void;
}

static int process_path( string dir )
{
   long hfile;
   struct _finddata_t de;
   string path;
   path = dir + "/*.*";
   if ( ( hfile = _findfirst( path.c_str(), &de ) ) == -1 ) {
      cout << "cannot open directory [" << dir << "]\n";
      return 0;
   }
   do {    // load and check all the *.arr and *.arr.gz files
      string file = de.name;
      if( !(( file == "." )||(file == "..")) ) {
         string full_path = dir + "/" + file;
         if( de.attrib & _A_SUBDIR ) {
            process_path( full_path ); // process the directory
         } else {
            // a file - check the extent
            int pos = file.find("."); // find first "."
            if( pos != -1 ) { // if found
               string f_index = file.substr(0, pos); // get index
               string ext = file.substr(pos + 1); // and extension
               if ( (ext == "arr") || (ext == "arr.gz") ) {
                  cout << file << "  " << f_index << "  '" << ext << "'" << endl;
                  process_array_file( full_path );
               } else {
                  cout << "Skipping [" << file << "]..." << endl;
               }
            } else {
               cout << "Skipping [" << file << "]..." << endl;
            }
         }
      }
   } while ( _findnext( hfile, &de ) == 0 );
   _findclose(hfile);
   return 1;
}

int main(int argc, char **argv) {

   sglog().setLogLevels( SG_ALL, SG_WARN );

   check_args( argc, argv );
   if ( argc != 2 ) {
      give_help( argv[0] );
      exit(1);
   }
   string src_array_path = argv[1];
   int   is_dir = is_file_or_directory( src_array_path );
   if(is_dir == 1)
      process_path( src_array_path );
   else if(is_dir == 2)
      process_array_file( src_array_path );
   else {
      cout << "Error: [ " << src_array_path << "] is neither a file or a directory!" << endl;
      return 1;
   }
   show_missing();
   return 0;
}

// eof - test4voids.cxx

